#ifndef BASE_SCATTERER_H
#define BASE_SCATTERER_H

  #include "Definitions.cpp"
  #include "utils.cpp"
  #include "Sources.cpp"
  #include "FibonnaciMesh.cpp"


class ScatteringProperties
{
public:
    size_t MaxOrder;
    double SizeParameter, Area;
    SOURCE::State source;

    virtual std::tuple<CVector, CVector> S1S2(const DVector &Phi){};
    virtual double GetQsca(){};
    virtual double GetQext(){};
    virtual double GetQback(){};
    virtual double GetG(){};


    ScatteringProperties(){}
    ScatteringProperties(double &Wavelength, CVector &Jones, double &Amplitude) : source(Wavelength, Jones, Amplitude){}
    ScatteringProperties(SOURCE::State &source) : source(source){}


    double GetQforward() {return GetQsca() - GetQback();};
    double GetQpr()      {return GetQext() - GetG() * GetQsca();};
    double GetQratio()   {return GetQback() / GetQsca();};
    double GetQabs()     {return GetQext() - GetQsca();};


    double  GetCsca()    {return GetQsca() * Area;};
    double  GetCext()    {return GetQext() * Area;};
    double  GetCabs()    {return GetQabs() * Area;};
    double  GetCback()   {return GetQback() * Area;};
    double  GetCforward(){return GetQforward() * Area;};
    double  GetCpr()     {return GetQpr() * Area;};
    double  GetCratio()  {return GetQratio() * Area;};

     DVector ComputePrefactor()
     {
       DVector Output; Output.reserve(MaxOrder);

       for (size_t m = 0; m < MaxOrder ; ++m)
          Output[m] = (double) ( 2 * (m+1) + 1 ) / ( (m+1) * ( (m+1) + 1 ) );

        return Output;
     }

    complex128 GetPropagator(const double &R)
    {
      return source.Amplitude / (source.k * R) * exp(-JJ*source.k*R);
    }

    std::tuple<CVector, CVector> ComputeFields_(CVector& S1, CVector& S2, DVector& Theta, double& R)
    {
      CVector EPhi, ETheta;
      size_t Size = Theta.size()*S1.size();

      EPhi.reserve(Size);
      ETheta.reserve(Size);

      complex128 propagator = GetPropagator(R);

      for (uint p=0; p < S1.size(); p++ )
          for (uint t=0; t < Theta.size(); t++ )
          {
            EPhi.push_back(   propagator * S1[p] * ( source.Jones[0] * cos(Theta[t]) + source.Jones[1] * sin(Theta[t]) ) );

            ETheta.push_back( propagator * S2[p] * ( source.Jones[0] * sin(Theta[t]) - source.Jones[1] * cos(Theta[t]) ) );
          }

      return std::make_tuple(EPhi, ETheta);
    }



    std::tuple<CVector, CVector> Fields_(const DVector& Phi, const DVector& Theta, const double R=1.0)
    {

      complex128 propagator = GetPropagator(R);

      auto [CosTerm, SinTerm] = PolarizationTerm(Theta);

      auto [S1, S2] = S1S2(Phi);

      auto EPhi = Unstructured(S2, CosTerm, - propagator);

      auto ETheta = Unstructured(S1, SinTerm, JJ * propagator);

      return std::make_tuple(EPhi, ETheta);
    }

    std::tuple<CVector, CVector> Fields_(const FibonacciMesh& Mesh, const double R=1.0)
    {
      return Fields_(Mesh.SCoord.Phi, Mesh.SCoord.Theta, R);
    }


    std::tuple<CVector, CVector, DVector, DVector> FullFields_(size_t& Sampling, double& R)
    {
      FullSteradian FullMesh = FullSteradian(Sampling);

      auto [S1, S2] = S1S2(FullMesh.SCoord.Phi);

      auto [EPhi, ETheta] = ComputeFields_(S1, S2, FullMesh.SCoord.Theta, R);

      return std::make_tuple(EPhi, ETheta, FullMesh.SCoord.Phi, FullMesh.SCoord.Theta);
    }


    std::tuple<DVector, DVector> PolarizationTerm(const DVector &Theta)
    {
      size_t  Sampling = Theta.size();

      DVector CosTerm, SinTerm;

      CosTerm.reserve(Sampling);
      SinTerm.reserve(Sampling);

      for (auto theta : Theta)
      {
          CosTerm.push_back( cos(theta) );
          SinTerm.push_back( sin(theta) );
      }
      return std::make_tuple(CosTerm, SinTerm);
    }



    std::tuple<CVector, FullSteradian> GetFullSPF(size_t Sampling, double R=1.0)
    {
      FullSteradian Mesh = FullSteradian(Sampling);

      auto [EPhi, ETheta] = StructuredFields(Mesh.SCoord.Phi, Mesh.SCoord.Theta, R);

      Squared(EPhi);
      Squared(ETheta);

      CVector spf = Add(EPhi, ETheta);

      return std::make_tuple(spf, Mesh);
    }



    std::tuple<CVector, CVector> StructuredFields(const DVector& Phi, const DVector& Theta, const double R)
    {
      complex128 propagator = GetPropagator(R);

      auto [CosTerm, SinTerm] = PolarizationTerm(Theta);

      auto [S1, S2] = S1S2(Phi);

      auto EPhiV      = Structured_(S2, CosTerm, source.Jones[0]*propagator);
      auto EThetaV    = Structured_(S1, SinTerm, source.Jones[1]*propagator);

      return std::make_tuple(EPhiV, EThetaV);
    }

    //--------------------------------------------------------PYTHON-------------------


    std::tuple<Cndarray,Cndarray> PyS1S2_(const DVector &Phi)
    {
      auto [S1, S2] = S1S2(Phi);

      return std::make_tuple( Vector2ndarray(std::move(S1)), Vector2ndarray(std::move(S2)))  ;
    }


    std::tuple<Cndarray, Cndarray, ndarray, ndarray> PyFullFields_(size_t &Sampling, double& R)
    {
      auto [EPhi, ETheta, Theta, Phi] = FullFields_(Sampling, R);

      Cndarray PyEPhi   = Vector2ndarray( std::move(EPhi),   {Sampling, Sampling} );
      Cndarray PyETheta = Vector2ndarray( std::move(ETheta), {Sampling, Sampling} );
      ndarray  PyTheta  = Vector2ndarray( std::move(Theta) );
      ndarray  PyPhi    = Vector2ndarray( std::move(Phi) );

      PyEPhi   = PyEPhi.attr("transpose")();
      PyETheta = PyETheta.attr("transpose")();

      return std::make_tuple(PyEPhi, PyETheta, PyPhi, PyTheta);
    }

    std::tuple<Cndarray,Cndarray> PyFields_(const DVector& Phi, const DVector& Theta, const double R)
    {
      auto [ETheta, EPhi] = Fields_(Phi, Theta, R);

      return std::make_tuple(Vector2ndarray(std::move(EPhi)), Vector2ndarray(std::move(ETheta)))  ;
    }



    double GetGWithFields_(size_t Sampling, double R)
    {
        auto [SPF, Mesh]      = GetFullSPF(Sampling, R);

        double Norm           = abs( Mesh.Integral(SPF) );
        double ExpectedCos    = abs( Mesh.IntegralCos(SPF) );

        return ExpectedCos/Norm;
    }

};

#endif
