#ifndef FIBONACCI_H
#define FIBONACCI_H

  #include "Coordinates.cpp"
  #include "numpy_interface.cpp"


  class FibonacciMesh
  {

    public:
      size_t      Sampling, TrueSample;
      Spherical   SCoord;
      Cartesian   CCoord, BaseCCoord;
      VectorField Perp, Para;
      DVector     HPara, VPara, HPerp, VPerp;
      double      MaxAngle, PhiOffset, GammaOffset, dOmega, Omega;

      VectorField VVector = VectorField({1,0,0}),
                  HVector = VectorField({0,1,0});

      void     mx_rot_y(double theta),
               mx_rot_x(double theta),
               mx_rot_z(double theta),
               mx_apply(Matrix3 M),
               GenCoordinate(),
               Cart2Sph();

      ndarray GetPara() {return Para.GetNumpy();};
      ndarray GetPerp() {return Perp.GetNumpy();};

      ndarray GetHPara() {return Vector2ndarrayCopy(HPara);};
      ndarray GetVPara() {return Vector2ndarrayCopy(VPara);};
      ndarray GetHPerp() {return Vector2ndarrayCopy(HPerp);};
      ndarray GetVPerp() {return Vector2ndarrayCopy(VPerp);};

      ndarray Getx() {return CCoord.GetPyX();};
      ndarray Gety() {return CCoord.GetPyY();};
      ndarray Getz() {return CCoord.GetPyZ();};


      ndarray GetR() {return SCoord.GetPyR();};
      ndarray GetPhi() {return SCoord.GetPyPhi();};
      ndarray GetTheta() { return SCoord.GetPyTheta();};

      FibonacciMesh(){}

      FibonacciMesh(const int    &Sampling,
                    const double &MaxAngle,
                    const double &PhiOffset,
                    const double &GammaOffset)
                      : Sampling(Sampling),
                        MaxAngle(MaxAngle),
                        PhiOffset(PhiOffset),
                        GammaOffset(GammaOffset)
      {

        CCoord = Cartesian(Sampling);
        GenMesh();
        BaseCCoord = CCoord;
        Rotations();
        SCoord = CCoord.Cart2Sph() ;
        ComputeVectorField();
        ComputeHVProjection();
      }


  void Rotations()
  {
    if (GammaOffset != 0.)
    {
      CCoord.mx_rot_x(GammaOffset);
      VVector.mx_rot_x(GammaOffset);
      HVector.mx_rot_x(GammaOffset);
    }

    if (PhiOffset   != 0.)
    {
      CCoord.mx_rot_y(PhiOffset);
      VVector.mx_rot_y(PhiOffset);
      HVector.mx_rot_y(PhiOffset);
    }
  }


  void ComputeVectorField()
  {
      Para = VectorField(Sampling);
      Perp = VectorField(Sampling);

      for (size_t i = 0; i < Sampling; i++)
      {
        Perp(i,0) = -CCoord.Y[i] ;
        Perp(i,1) = CCoord.X[i] ;
        Perp(i,2) = 0.0;

        Para(i,0) = CCoord.X[i] * CCoord.Z[i] ;
        Para(i,1) = CCoord.Y[i] * CCoord.Z[i] ;
        Para(i,2) = -( pow( CCoord.X[i], 2 ) + pow( CCoord.Y[i], 2 ) );
      }
    Para.Normalize();
    Perp.Normalize();
  }




  void ComputeHVProjection()
  {
    HPara = Para.ScalarProduct(HVector);
    VPara = Para.ScalarProduct(VVector);

    HPerp = Perp.ScalarProduct(HVector);
    VPerp = Perp.ScalarProduct(VVector);
  }


  double NA2Angle(const double &NA)
  {
    if (NA <= 1.0)
        return asin(NA);

    if (NA >= 1.0)
        return asin(NA-1.0) + PI/2.0;

    return 1.0;
  }


  void GenMesh()
  {
      this->GetProperties();

      double phi  = PI * (3. - sqrt(5.));  // golden angle = 2.39996322972865332

      for (size_t i = 0; i < TrueSample; i++){
         double theta    = phi * i;
         CCoord.Z[i]     = 1 - (i / (double)( TrueSample - 1) ) * 2 ;
         double radius   = sqrt(1 - CCoord.Z[i] * CCoord.Z[i]);
         CCoord.X[i]     = cos(theta) * radius ;
         CCoord.Y[i]     = sin(theta) * radius ;

         if (i == (size_t)Sampling-1) break;
        }
  }


  void GetProperties()
  {
    double SolidAngle   = abs( 2. * PI * ( cos(MaxAngle) - 1. ) ),   //cos(0) =1
           ratio        = ( 4. * PI / SolidAngle );

    size_t _TrueSample   = (size_t) ( Sampling * ratio );

    dOmega     = 4. * PI / _TrueSample;
    Omega      = dOmega * Sampling ;
    TrueSample = _TrueSample;
  }

};



class FullSteradian
{
  public:
    size_t   Sampling;

    Spherical SCoord;

    double dTheta, dPhi;

    FullSteradian(size_t Sampling) : Sampling(Sampling), SCoord(Sampling)
    {
      dTheta = 2.0 * PI / (Sampling-1);
      dPhi   = 1.0 * PI / (Sampling-1);

      for (size_t p=0; p<Sampling; p++)
          SCoord.Phi[p]   = p * dPhi   - PI/2.0;

      for (size_t t=0; t<Sampling; t++)
          SCoord.Theta[t] = t * dTheta - PI/1.0;


    }

    template<typename T> T Integral(std::vector<T>& Vector)
    {
      T integral = 0;

      for (size_t p=0; p<Sampling; p++)
          for (size_t t=0; t<Sampling; t++)
            integral += Vector[p*Sampling + t] * sin(SCoord.Phi[p] + PI/2.0) * dPhi * dTheta;

      return integral;
    }


    template<typename T> T IntegralCos(std::vector<T>& Vector)
    {
      T integral = 0;

      for (size_t p=0; p<Sampling; p++)
          for (size_t t=0; t<Sampling; t++)
            integral += Vector[p*Sampling + t] * cos(SCoord.Phi[p] + PI/2.0) * sin(SCoord.Phi[p] + PI/2.0) * dPhi * dTheta;

      return integral;
    }

    double Integral()
    {
      double integral = 0;

      for (auto phi : SCoord.Phi)
          for (auto theta : SCoord.Theta)
            integral += sin(phi+PI/2.0) * dPhi * dTheta;

      return integral;
    }
};


#endif

// --
