#ifndef SPHERE_H
#define SPHERE_H

  #include "BaseScatterer.cpp"
  #include "numpy_interface.cpp"
  #include "VSH.cpp"
  #include <optional>

  namespace SPHERE
  {
    struct State
    {
      double Diameter;
      complex128 Index;
      double nMedium;
      State(){}
      State(double &Diameter, complex128 &Index, double &nMedium)
                  : Diameter(Diameter), Index(Index), nMedium(nMedium){}

      State(double &Diameter, complex128 &&Index, double &nMedium)
                  : Diameter(Diameter), Index(Index), nMedium(nMedium){}

      void ApplyMedium()
      {
        this->Index    /= this->nMedium;
        this->Diameter *= this->nMedium;
      }
    };

    class Scatterer: public ScatteringProperties
    {

        public:
            std::optional<CVector> _An, _Bn, _Cn, _Dn;

            State     state;

            Cndarray GetPyAn() { return Vector2ndarray(_An.value(), {MaxOrder}); }
            Cndarray GetPyBn() { return Vector2ndarray(_Bn.value(), {MaxOrder}); }
            Cndarray GetPyCn() { return Vector2ndarray(_Cn.value(), {MaxOrder}); }
            Cndarray GetPyDn() { return Vector2ndarray(_Dn.value(), {MaxOrder}); }

            CVector GetAn() { if (_An) return _An.value(); else{ ComputeAnBn(); return _An.value(); } };
            CVector GetBn() { if (_Bn) return _Bn.value(); else{ ComputeAnBn(); return _Bn.value(); } };
            CVector GetCn() { if (_An) return _Cn.value(); else{ ComputeCnDn(); return _Cn.value(); } };
            CVector GetDn() { if (_Bn) return _Dn.value(); else{ ComputeCnDn(); return _Dn.value(); } };

            Scatterer(){}
            Scatterer(double     &Wavelength,
                      double     &Amplitude,
                      double     &Diameter,
                      complex128 &Index,
                      double     &nMedium,
                      CVector    &Jones,
                      size_t      MaxOrder = 0 )
                      : state(Diameter, Index, nMedium), ScatteringProperties(Wavelength, Jones, Amplitude)
                      {
                        this->MaxOrder = MaxOrder;
                        Init();
                      }

            Scatterer(State &state, SOURCE::State &source, size_t MaxOrder = 0)
                      : state(state), ScatteringProperties(source)
                      {
                        this->MaxOrder = MaxOrder;
                        Init();
                      }

            void Init()
            {
              state.ApplyMedium();
              this->SizeParameter = PI * state.Diameter / source.Wavelength;
              if (MaxOrder == 0)
                  this->MaxOrder = (size_t) (2 + this->SizeParameter + 4 * pow(this->SizeParameter, 1./3.)) + 16;
              this->Area = PI * pow(state.Diameter/2.0, 2);
              ComputeAnBn();

            }


        void ComputeAnBn()
        {
          _An = CVector(MaxOrder);
          _Bn = CVector(MaxOrder);

          complex128 x   = SizeParameter,
                     m   = state.Index,
                     mx  = m * x,
                     _da, _db, _gsx, _gs1x, _px, _chx, _p1x, _ch1x, _p2x, _ch2x;

          size_t     nmx = std::max( MaxOrder, (size_t) std::abs(mx) ) + 16;
          CVector    Dn  = VSH::SPHERICAL::ComputeDn(nmx, mx);

          double     n;

          _p1x  = sin(x);
          _ch1x = cos(x);

          for (size_t i = 1; i < MaxOrder+1; ++i)
            {
                n          = (double) i;
                _px        =  x * F90jn(n, x);
                _chx       = -x * F90yn(n, x);

                _p2x       = _px;
                _ch2x      = _chx;

                _gsx       =  _px  - 1.*JJ * _chx;
                _gs1x      =  _p1x - 1.*JJ * _ch1x;

                _da        = Dn[i]/m + n/x ;
                _db        = Dn[i]*m + n/x ;

                _An.value()[i-1] = (_da * _px - _p1x) / (_da * _gsx - _gs1x) ;
                _Bn.value()[i-1] = (_db * _px - _p1x) / (_db * _gsx - _gs1x) ;

                _p1x  = _p2x;
                _ch1x = _ch2x;
            }
        }


      void ComputeCnDn(size_t MaxOrder = 0)
      {

        if (MaxOrder == 0) MaxOrder = this->MaxOrder;

        _Cn = CVector(MaxOrder);
        _Dn = CVector(MaxOrder);

        complex128 x   = SizeParameter,
                   m   = state.Index,
                   z  = m * x;

        size_t nmx = std::max( MaxOrder, (size_t) std::abs(z) ) + 16;

        CVector Cnx = CVector(nmx),
                Cnn, jnx, jnmx, yx, hx, b1x, y1x, hn1x, ax, ahx, numerator,
                c_denominator, d_denominator;

        b1x.push_back( +sin(x) / x );
        y1x.push_back( -cos(x) / x );

        for (double i = nmx; i > 1; i--)
        {
          Cnx[i-2] = i - z*z/(Cnx[i-1] + i);
        }

        for (size_t i = 0; i < MaxOrder; i++)
        {
          double n = (double) i;
          Cnn.push_back( Cnx[i] );
          jnx.push_back(  F90jn( n+1, x ) );

          jnmx.push_back( 1. / ( F90jn(n+1, z ) ) );
          yx.push_back( F90yn(n+1, x ) );
          hx.push_back( jnx[i] + JJ * yx[i] );

          b1x.push_back( jnx[i] );
          y1x.push_back( yx[i] );
          hn1x.push_back( b1x[i] + JJ * y1x[i] );

          ax.push_back(  x * b1x[i]  - ( n+1 ) * jnx[i] );
          ahx.push_back( x * hn1x[i] - ( n+1 ) * hx[i]  );

          numerator.push_back( jnx[i] * ahx[i] - hx[i] * ax[i] );
          c_denominator.push_back( ahx[i] - hx[i] * Cnn[i] );
          d_denominator.push_back( m * m * ahx[i] - hx[i] * Cnn[i] );
          _Cn.value()[i] = jnmx[i] * numerator[i] / c_denominator[i] ;
          _Dn.value()[i] = jnmx[i] * m * numerator[i] / d_denominator[i] ;
        }
      }


      double GetG()
      {
          double temp = 0;

          CVector an = GetAn(), bn = GetBn();

          for(size_t it = 0; it < MaxOrder-1; ++it)
          {
              double n = (double) it + 1;
              temp += ( n * (n + 2.) / (n + 1.) ) * std::real(an[it] * std::conj(an[it+1]) + bn[it] * std::conj(bn[it+1]) );
              temp += ( (2. * n + 1. ) / ( n * (n + 1.) ) )  * std::real( an[it] * std::conj(bn[it]) );
          }

          return temp * 4. / ( GetQsca() * pow(SizeParameter, 2) );
      }



    std::tuple<CVector, CVector> S1S2(const DVector &Phi)
    {
      CVector an = GetAn(), bn = GetBn();

      CVector S1(Phi.size(), 0.0), S2(Phi.size(), 0.0);

      DVector Prefactor = ComputePrefactor();

      DVector Mu; Mu.reserve(Phi.size());

      for (double phi : Phi)
          Mu.push_back( cos( phi-PI/2.0 ) );


      for (uint i = 0; i < Phi.size(); i++)
      {
          auto [pin, taun] = VSH::SPHERICAL::MiePiTau( Mu[i], MaxOrder);

          for (uint m = 0; m < MaxOrder ; m++){
              S1[i]    += Prefactor[m] * ( an[m] * pin[m] +  bn[m] * taun[m] );
              S2[i]    += Prefactor[m] * ( an[m] * taun[m] + bn[m] * pin[m]  );
            }
      }

      return std::make_tuple(S1, S2)  ;
    }



    double GetQsca()
    {
        CVector an = GetAn(), bn = GetBn();

        double temp = 0;
        for(size_t it = 0; it < MaxOrder; ++it)
        {
             double n = (double) it + 1;
             temp += (2.* n + 1.) * ( pow( std::abs(an[it]), 2) + pow( std::abs(bn[it]), 2)  );

        }
        return temp * 2. / pow( SizeParameter, 2.);
    }


    double GetQext()
    {
      CVector an = GetAn(), bn = GetBn();

      double temp = 0;
      for(size_t it = 0; it < MaxOrder; ++it)
      {
           double n = (double) it + 1;
           temp += (2.* n + 1.) * std::real( an[it] + bn[it] );

      }
      return temp * 2. / pow( SizeParameter, 2.);
    }


    double GetQback()
    {
        CVector an = GetAn(), bn = GetBn();

        complex128 temp = 0;
        for(size_t it = 0; it < MaxOrder-1; ++it)
        {
          double n = (double) it + 1;
          temp += (2. * n + 1) * pow(-1., n) * ( an[it] - bn[it] ) ;
        }

        temp *= pow( std::abs(temp), 2. ) / pow( SizeParameter, 2. );
        return std::abs(temp);
    }



  };

}


#endif
// -
